from collections import defaultdict
from ..utils.thread import ThreadLocalObject


class NodeHistory(ThreadLocalObject):

    def __init__(self):
        super(NodeHistory, self).__init__()
        self._history = defaultdict(list)

    def __len__(self):
        return sum(len(nodes) for nodes in self._history.values())

    def log(self, method, *nodes):
        self._history[method].extend(nodes)

    def list(self, method):
        return self._history[method]

    def clear(self):
        self._history.clear()



from unittest2 import TestCase
from contextlib import contextmanager


class BaseTest(TestCase):

    def setUp(self):
        from cio.environment import env
        from cio.backends import cache, storage
        from cio.pipeline import pipeline
        from cio.plugins import plugins

        env.reset()
        cache.clear()
        storage.backend._call_delete()
        pipeline.clear()
        plugins.load()

        self.configure()

    def configure(self):
        from cio.conf import settings
        settings.configure(
            ENVIRONMENT={
                'default': {
                    'i18n': 'sv-se',
                    'l10n': 'tests',
                    'g11n': 'global'
                }
            },
        )

    def assertKeys(self, dict, *keys):
        self.assertEqual(set(dict.keys()), set(keys))

    @contextmanager
    def assertCache(self, calls=-1, hits=-1, misses=-1, sets=-1):
        from cio.backends import cache

        cb = cache.backend

        cb.calls = 0
        cb.hits = 0
        cb.misses = 0
        cb.sets = 0

        yield

        if calls >= 0:
            assert cb.calls == calls
        if hits >= 0:
            assert cb.hits == hits
        if misses >= 0:
            assert cb.misses == misses
        if sets >= 0:
            assert cb.sets == sets


    @contextmanager
    def assertDB(self, calls=-1, selects=-1, inserts=-1, updates=-1, deletes=-1):
        from cio.backends import storage
        backend = storage.backend
        backend.start_debug()

        yield

        count = lambda cmd: len([q for q in backend.queries if q['sql'].split(' ', 1)[0].upper().startswith(cmd)])
        if calls >= 0:
            assert len(backend.queries) == calls
        if selects >= 0:
            assert count('SELECT') == selects
        if inserts >= 0:
            assert count('INSERT') == inserts
        if updates >= 0:
            assert count('UPDATE') == updates
        if deletes >= 0:
            assert count('DELETE') == deletes

        backend.stop_debug()



from cio.backends import get_backend, storage
from cio.backends.base import CacheBackend, StorageBackend, DatabaseBackend
from cio.backends.exceptions import InvalidBackend, PersistenceError, NodeDoesNotExist
from cio.backends.sqlite import SqliteBackend
from cio.conf.exceptions import ImproperlyConfigured
from cio.utils.uri import URI
from tests import BaseTest


class StorageTest(BaseTest):

    def test_get_storage(self):
        backend = get_backend('sqlite://:memory:')
        self.assertTrue(issubclass(backend.__class__, StorageBackend))

        backend = '.'.join((SqliteBackend.__module__, SqliteBackend.__name__))
        with self.assertRaises(ImproperlyConfigured):
            backend = get_backend(backend)

        backend = get_backend({
            'BACKEND': backend,
            'NAME': ':memory:'
        })
        self.assertTrue(issubclass(backend.__class__, DatabaseBackend))

        with self.assertRaises(ImportError):
            get_backend('foo.Bar')
        with self.assertRaises(ImportError):
            get_backend('cio.storage.backends.orm.Bogus')
        with self.assertRaises(InvalidBackend):
            get_backend('invalid')
        with self.assertRaises(InvalidBackend):
            get_backend('foo://')

    def test_bogus_backend(self):
        class BogusStorage(CacheBackend, StorageBackend):
            pass
        bogus = BogusStorage()
        self.assertIsNone(bogus.scheme)
        with self.assertRaises(NotImplementedError):
            bogus._get(None)
        with self.assertRaises(NotImplementedError):
            bogus._get_many(None)
        with self.assertRaises(NotImplementedError):
            bogus._set(None, None)
        with self.assertRaises(NotImplementedError):
            bogus._set_many(None)
        with self.assertRaises(NotImplementedError):
            bogus._delete(None)
        with self.assertRaises(NotImplementedError):
            bogus._delete_many(None)
        with self.assertRaises(NotImplementedError):
            bogus.publish(None)
        with self.assertRaises(NotImplementedError):
            bogus.get_revisions(None)

    def test_create_update(self):
        storage.set('i18n://sv-se@a.txt#draft', u'first')
        node = storage.get('i18n://sv-se@a#draft')
        self.assertEqual(node['content'], u'first')
        self.assertEqual(node['uri'], 'i18n://sv-se@a.txt#draft')
        storage.set('i18n://sv-se@a.txt#draft', u'second')
        node = storage.get('i18n://sv-se@a#draft')
        self.assertEqual(node['content'], u'second')
        self.assertEqual(node['uri'], 'i18n://sv-se@a.txt#draft')

    def test_get(self):
        storage.set('i18n://sv-se@a.txt#draft', u'A')
        storage.set('i18n://sv-se@b.md#draft', u'B')
        node = storage.get('i18n://sv-se@a#draft')
        self.assertEqual(node['uri'], 'i18n://sv-se@a.txt#draft')
        self.assertEqual(node['content'], u'A')

        storage.publish('i18n://sv-se@a#draft')
        storage.publish('i18n://sv-se@b#draft')

        nodes = storage.get_many(('i18n://sv-se@a', 'i18n://sv-se@b'))
        for node in nodes.values():
            node.pop('meta')
        self.assertDictEqual(nodes, {
            'i18n://sv-se@a': {
                'uri': 'i18n://sv-se@a.txt#1',
                'content': u'A'
            },
            'i18n://sv-se@b': {
                'uri': 'i18n://sv-se@b.md#1',
                'content': u'B'
            }
        })

    def test_delete(self):
        storage.set('i18n://sv-se@a.txt#draft', u'A')
        storage.set('i18n://sv-se@b.txt#draft', u'B')

        node = storage.get('i18n://sv-se@a#draft')
        self.assertEqual(node['content'], u'A')

        deleted_node = storage.delete('sv-se@a#draft')
        deleted_node.pop('meta')
        self.assertDictEqual(deleted_node, {'uri': 'i18n://sv-se@a.txt#draft', 'content': u'A'})

        deleted_nodes = storage.delete_many(('sv-se@a#draft', 'sv-se@b#draft'))
        for node in deleted_nodes.values():
            node.pop('meta')
        self.assertDictEqual(deleted_nodes, {
            'i18n://sv-se@b#draft': {'uri': 'i18n://sv-se@b.txt#draft', 'content': u'B'}
        })

    def test_nonexisting_node(self):
        with self.assertRaises(URI.Invalid):
            storage.get('?')
        with self.assertRaises(NodeDoesNotExist):
            storage.get('sv-se@page/title')

    def test_plugin_mismatch(self):
        storage.set('i18n://sv-se@a.txt#draft', u'A')
        storage.publish('i18n://sv-se@a.txt#draft')

        with self.assertRaises(NodeDoesNotExist):
            storage.get('i18n://sv-se@a.md')

        nodes = storage.get_many(('i18n://sv-se@a.md',))
        self.assertDictEqual(nodes, {})

    def test_node_integrity(self):
        storage.backend._create(URI('i18n://sv-se@a.txt#draft'), u'first')
        with self.assertRaises(PersistenceError):
            storage.backend._create(URI('i18n://sv-se@a'), u'second')
        with self.assertRaises(PersistenceError):
            storage.backend._create(URI('i18n://sv-se@a.txt'), u'second')
        with self.assertRaises(PersistenceError):
            storage.backend._create(URI('i18n://sv-se@a#draft'), u'second')



from .txt import TextPlugin


class MarkdownPlugin(TextPlugin):

    ext = 'md'

    def render(self, data):
        # TODO: Handle markdown import error
        import markdown
        if data:
            extensions = self.settings.get('EXTENSIONS', [])
            return markdown.markdown(data, extensions=extensions)



from collections import defaultdict
from ..node import Node
from ..utils.thread import ThreadLocalObject


class BufferedNode(Node):

    def __init__(self, node, callback):
        self._node = node
        self._callback = callback
        self._flushed = False

    def __repr__(self):
        return u'<BufferedNode: %s>' % self.uri

    def flush(self):
        if not self._flushed:
            self._callback()

    @property
    def uri(self):
        self.flush()
        return self._node.uri

    def get_content(self):
        self.flush()
        return self._node.content

    def set_content(self, content):
        self._flushed = True
        self._node.content = content

    content = property(get_content, set_content)

    @property
    def meta(self):
        return self._node.meta

    @property
    def initial(self):
        return self._node.initial

    @property
    def initial_uri(self):
        return self._node.initial_uri


class NodeBuffer(ThreadLocalObject):

    def __init__(self):
        super(NodeBuffer, self).__init__()
        self._buffer = {}

    def __len__(self):
        return sum(len(method_nodes) for method_nodes in self._buffer.values())

    def add(self, method, node):
        if method not in self._buffer:
            self._buffer[method] = defaultdict(list)

        buffer = self._buffer[method]
        buffer[node.initial_uri].append(node)

    def pop(self, method):
        buffered_nodes = dict(self._buffer[method])
        self._buffer[method].clear()
        return buffered_nodes

    def clear(self):
        self._buffer.clear()



import json
import cio
from cio.conf import settings
from cio.plugins import plugins
from cio.backends import storage
from cio.plugins.base import BasePlugin
from cio.plugins.exceptions import UnknownPlugin
from cio.plugins.txt import TextPlugin
from tests import BaseTest


class PluginTest(BaseTest):

    def test_resolve_plugin(self):
        with self.assertRaises(UnknownPlugin):
            plugins.get('xyz')

        plugin = plugins.resolve('i18n://sv-se@page/title.txt')
        self.assertIsInstance(plugin, TextPlugin)

        with self.assertRaises(UnknownPlugin):
            plugins.resolve('i18n://sv-se@page/title.foo')

    def test_register_plugin(self):
        with self.assertRaises(ImportError):
            plugins.register('foo.bar.BogusPlugin')
        with self.assertRaises(ImportError):
            plugins.register('cio.plugins.text.BogusPlugin')

    def test_plugin(self):
        self.assertSetEqual(set(plugins.plugins.keys()), set(['txt', 'md']))

        settings.configure(PLUGINS=[
            'cio.plugins.txt.TextPlugin',
            'cio.plugins.md.MarkdownPlugin',
            'tests.test_plugins.UppercasePlugin'
        ])

        self.assertSetEqual(set(plugins.plugins.keys()), set(['txt', 'md', 'up']))

        node = cio.set('sv-se@page/title.up', {'name': u'lundberg'}, publish=False)
        self.assertListEqual(node._content, [{'name': u'lundberg'}, u'{"name": "lundberg"}', u'LUNDBERG'])

        cio.publish(node.uri)

        node = cio.get('page/title')
        raw_content = storage.get(node.uri)

        self.assertIsNotNone(raw_content)
        self.assertEqual(raw_content['uri'], 'i18n://sv-se@page/title.up#1')
        self.assertEqual(raw_content['content'], u'{"name": "lundberg"}')
        self.assertEqual(node.content, u'LUNDBERG')
        self.assertEqual(node.uri.ext, 'up')

        self.assertSetEqual(set(p for p in plugins), set(('txt', 'md', 'up')))


class UppercasePlugin(BasePlugin):

    ext = 'up'

    def load(self, content):
        return json.loads(content) if content else None

    def save(self, data):
        return json.dumps(data)

    def render(self, data):
        return data['name'].upper()



from .environment import env
from .utils.uri import URI

empty = object()


class Node(object):

    def __init__(self, uri, content=None, **meta):
        self.env = env.state
        self._uri = [uri, URI(uri)]
        self._content = [content]
        self.meta = meta

    def __repr__(self):
        return u'<Node: %s>' % self.uri

    def __str__(self):
        content = self.render()
        if isinstance(content, unicode):
            content = content.encode('utf-8')
        return content or ''

    def __unicode__(self):
        return self.render() or u''

    def render(self, **context):
        if self.content is not None:
            if context:
                return self.content.format(**context)
            else:
                return self.content

    def get_uri(self):
        return self._uri[-1]

    def set_uri(self, uri):
        if uri != self.get_uri():
            self._uri.append(URI(uri))

    uri = property(get_uri, set_uri)

    def get_content(self):
        return self._content[-1]

    def set_content(self, content):
        if content != self.get_content():
            self._content.append(content)

    content = property(get_content, set_content)

    @property
    def initial(self):
        return self._content[0]

    @property
    def initial_uri(self):
        return self._uri[0]

    def for_json(self):
        return {
            'uri': str(self.uri),
            'content': self.content,
            'meta': self.meta if self.meta is not None else {}
        }



# coding=utf-8
from cio.utils.uri import URI
from tests import BaseTest


class UtilsTest(BaseTest):

    def test_uri(self):
        self.assertEqual(URI('i18n://sv@page/title.txt'), 'i18n://sv@page/title.txt')

        uri = URI(scheme='i18n', namespace='sv-se', path='page/title', ext='txt')
        self.assertEqual(uri, 'i18n://sv-se@page/title.txt')

        uri = URI('page/title')
        self.assertFalse(uri.is_absolute())
        self.assertEqual(uri.scheme, 'i18n')
        self.assertIsNone(uri.namespace)
        self.assertEqual(uri.path, 'page/title')
        self.assertIsNone(uri.ext)

        uri = uri.clone(namespace='sv-se')
        self.assertFalse(uri.is_absolute())
        self.assertEqual(uri, 'i18n://sv-se@page/title')

        uri = uri.clone(ext='txt')
        self.assertEqual(uri, 'i18n://sv-se@page/title.txt')
        self.assertTrue(uri.is_absolute())

        uri = uri.clone(path='images/me.jpg/title', version='draft')
        self.assertEqual(uri, 'i18n://sv-se@images/me.jpg/title.txt#draft')
        self.assertEqual(uri.version, 'draft')
        self.assertEqual(uri.path, 'images/me.jpg/title')
        self.assertEqual(uri.ext, 'txt')

        uri = uri.clone(ext=None)
        self.assertEqual(uri, 'i18n://sv-se@images/me.jpg/title#draft')
        self.assertIsNone(uri.ext)

        uri = URI('page/title')
        uri = uri.clone(scheme=None)
        self.assertEqual(uri, 'page/title')

        uri = URI('i18n://sv@page/title.txt#draft')
        self.assertEqual(uri, 'i18n://sv@page/title.txt#draft')
        self.assertEqual(uri.version, 'draft')



from .base import BasePipe
from ...conf import settings
from ...backends import storage
from ...backends.exceptions import NodeDoesNotExist


class StoragePipe(BasePipe):

    def get_request(self, request):
        response = {}
        stored_nodes = storage.get_many(request.keys())

        for uri, stored_node in stored_nodes.iteritems():
            node = response[node.uri] = request.pop(uri)
            self.materialize_node(node, **stored_node)

        return response

    def get_response(self, response):
        if response:
            # Redirect nodes without extension (non-persisted) to default
            for node in response.values():
                if not node.uri.ext:
                    node.uri = node.uri.clone(ext=settings.URI_DEFAULT_EXT)

        return response

    def set_request(self, request):
        for node in request.values():
            stored_node, _ = storage.set(node.uri, node.content, **node.meta)
            uri = stored_node['uri']
            node.uri = uri
            node.meta = stored_node['meta']

    def delete_request(self, request):
        deleted_nodes = storage.delete_many(request.keys())

        for uri, deleted_node in deleted_nodes.iteritems():
            node = request[uri]
            deleted_node['content'] = None  # Set content to None to signal node has been deleted
            self.materialize_node(node, **deleted_node)

    def publish_request(self, request):
        for uri, node in request.items():
            try:
                published_node = storage.publish(uri, **node.meta)
            except NodeDoesNotExist:
                request.pop(uri)
            else:
                node = request[uri]
                self.materialize_node(node, **published_node)


class NamespaceFallbackPipe(BasePipe):

    def get_request(self, request):
        response = {}
        fallback_uris = {}

        # Build fallback URI map
        for uri, node in request.iteritems():
            # if node.uri != node.initial_uri:
            namespaces = getattr(node.env, uri.scheme)[1:]
            if namespaces:
                uris = [uri.clone(namespace=namespace) for namespace in namespaces]
                fallback_uris[node.uri] = uris

        # Fetch nodes from storage, each fallback level slice at a time
        while fallback_uris:
            level_uris = dict((fallback.pop(0), uri) for uri, fallback in fallback_uris.iteritems())
            stored_nodes = storage.get_many(level_uris.keys())

            # Set node fallback content and add to response
            for uri, stored_node in stored_nodes.iteritems():
                requested_uri = level_uris[uri]
                node = response[node.uri] = request.pop(requested_uri)
                self.materialize_node(node, **stored_node)

            # Remove exhausted uris that has run out of fallback namespaces
            for uri, fallback in fallback_uris.items():
                if not fallback:
                    fallback_uris.pop(uri)

        return response



# coding=utf-8
from ..conf import settings


class URI(str):

    @staticmethod
    def __new__(cls, uri=None, scheme=None, namespace=None, path=None, ext=None, version=None):
        if isinstance(uri, URI):
            return uri
        elif uri is not None:
            return URI._parse(uri)
        else:
            return URI._render(scheme, namespace, path, ext, version)

    @classmethod
    def _parse(cls, uri):
        base, _, version = uri.partition(settings.URI_VERSION_SEPARATOR)
        scheme, _, path = base.rpartition(settings.URI_SCHEME_SEPARATOR)
        namespace, _, path = path.rpartition(settings.URI_NAMESPACE_SEPARATOR)
        _path, _, ext = path.rpartition(settings.URI_EXT_SEPARATOR)
        if '/' in ext:
            ext = ''
        else:
            path = _path

        if not path and ext:
            path, ext = ext, ''

        return cls._render(
            scheme or 'i18n',
            namespace or None,
            path,
            ext or None,
            version or None
        )

    @classmethod
    def _render(cls, scheme, namespace, path, ext, version):
        def parts_gen():
            if scheme:
                yield scheme
                yield settings.URI_SCHEME_SEPARATOR
            if namespace:
                yield namespace
                yield settings.URI_NAMESPACE_SEPARATOR
            if path:
                yield path
                if ext:
                    yield settings.URI_EXT_SEPARATOR
                    yield ext
                if version:
                    yield settings.URI_VERSION_SEPARATOR
                    yield version

        uri = str.__new__(cls, ''.join(parts_gen()))
        uri.scheme = scheme
        uri.namespace = namespace
        uri.path = path
        uri.ext = ext
        uri.version = version
        return uri

    def is_absolute(self):
        """
        Validates that uri contains all parts except version
        """
        return self.namespace and self.ext and self.scheme and self.path

    def has_parts(self, *parts):
        return not any(getattr(self, part, None) is None for part in parts)

    def clone(self, **parts):
        part = lambda part: parts.get(part, getattr(self, part))
        args = (part(p) for p in ('scheme', 'namespace', 'path', 'ext', 'version'))
        return URI._render(*args)

    class Invalid(Exception):
        pass



from collections import defaultdict
from functools import partial
from .buffer import NodeBuffer, BufferedNode
from .history import NodeHistory
from ..conf import settings
from ..utils.imports import import_class

PIPELINE_CALLS = ('get', 'set', 'delete', 'publish')


class PipelineHandler(object):

    def __init__(self):
        self.history = NodeHistory()
        self._buffer = NodeBuffer()
        self.load()
        settings.watch(self.load)

    def load(self):
        self.pipes = []
        for pipe_path in settings.PIPELINE:
            self.add_pipe(pipe_path)
        self.build()

    def add_pipe(self, pipe):
        try:
            if isinstance(pipe, type):
                pipe_class = pipe
            else:
                pipe_class = import_class(pipe)
        except ImportError as e:
            raise ImportError('Could not import content-io pipe "%s" (Is it on sys.path?): %s' % (pipe, e))
        else:
            self.pipes.append(pipe_class)

    def build(self):
        self._pipeline = defaultdict(list)
        for pipe_class in self.pipes:
            pipe = pipe_class()
            for call in PIPELINE_CALLS:
                request_handler = getattr(pipe, '%s_request' % call, None)
                response_handler = getattr(pipe, '%s_response' % call, None)
                self._pipeline[call].append((request_handler, response_handler))

    def send(self, method, *nodes):
        request = dict((node.uri, node) for node in nodes)
        response_chain = []

        # Iterate request chain
        for request_handler, response_handler in self._pipeline[method]:
            if request_handler:
                pipe_response = request_handler(request)
            else:
                pipe_response = None

            # Build response chain
            response_chain.append((response_handler, pipe_response))

            # Go no further in pipeline if out of node requests
            if not request:
                break

        # Turn request to response
        response = request

        # Iterate response chain
        for response_handler, pipe_response in reversed(response_chain):
            if response and response_handler:
                response = response_handler(response)
            if pipe_response:
                response.update(pipe_response)

        # Log response
        self.history.log(method, *response.values())

        return response

    def buffer(self, method, node):
        callback = partial(self.flush, method)
        buffered_node = BufferedNode(node, callback=callback)
        self._buffer.add(method, buffered_node)
        return buffered_node

    def flush(self, method):
        # Extract nodes from buffer
        buffer = self._buffer.pop(method)

        # Extract and flatten wrapped nodes, send only distinct uri's
        nodes = (buffered_nodes[0]._node for buffered_nodes in buffer.values())

        # Send nodes through pipeline
        response = self.send(method, *nodes)

        # Update buffered nodes to make sure uri duplicates, not sent through pipeline, gets content
        for node in response.values():
            for buffered_node in buffer[node.initial_uri]:
                buffered_node.content = node.content

    def clear(self):
        self._buffer.clear()
        self.history.clear()



from collections import namedtuple
from contextlib import contextmanager
from .conf import settings
from .utils.thread import ThreadLocalObject

DEFAULT = 'default'
State = namedtuple('State', ['i18n', 'l10n', 'g11n'])


class Environment(ThreadLocalObject):

    def __init__(self):
        super(Environment, self).__init__()
        self.reset()
        settings.watch(self.reset)

    @contextmanager
    def __call__(self, name=None, i18n=None, l10n=None, g11n=None):
        if name:
            self.push(name)
        else:
            self.push_state(i18n=i18n, l10n=l10n, g11n=g11n)
        yield
        self.pop()

    def reset(self):
        self._stack = []
        self.push(DEFAULT)

    def push(self, name):
        env = settings.ENVIRONMENT[name]
        self.push_state(**env)

    def _ensure_tuple(self, ns):
        if isinstance(ns, tuple):
            return ns
        elif hasattr(ns, '__iter__'):
            return tuple(ns)
        else:
            return ns,

    def push_state(self, i18n=None, l10n=None, g11n=None):
        i18n = self._ensure_tuple(i18n or self.i18n)
        l10n = self._ensure_tuple(l10n or self.l10n)
        g11n = self._ensure_tuple(g11n or self.g11n)
        state = State(i18n, l10n, g11n)
        self._stack.append(state)

    def pop(self):
        if len(self._stack) == 1:
            raise IndexError('Unable to pop last environment state')
        self._stack.pop()

    @property
    def state(self):
        return self._stack[-1]

    @property
    def i18n(self):
        return self.state.i18n

    @property
    def l10n(self):
        return self.state.l10n

    @property
    def g11n(self):
        return self.state.g11n

env = Environment()



import cio
from cio.backends import cache, storage
from cio.backends.exceptions import NodeDoesNotExist
from cio.utils.uri import URI
from tests import BaseTest


class CacheTest(BaseTest):

    uri = 'i18n://sv-se@label/email.txt'

    def test_cached_node(self):
        with self.assertRaises(NodeDoesNotExist):
            storage.get(self.uri)

        content = cache.get(self.uri)
        self.assertIsNone(content)

        node, _ = storage.set(self.uri + '#draft', u'e-post')
        storage.publish(node['uri'])

        with self.assertCache(calls=2, misses=1, sets=1):
            node = cio.get('i18n://label/email', lazy=False)

        cached_node = cache.get('i18n://sv-se@label/email')
        self.assertIsInstance(cached_node, dict)
        self.assertKeys(cached_node, 'uri', 'content')
        _uri, content = cached_node['uri'], cached_node['content']
        self.assertEqual(_uri, 'i18n://sv-se@label/email.txt#1')
        self.assertTrue(content == node.content == u'e-post')

        with self.assertCache(calls=1, misses=0, hits=1):
            node = cio.get('i18n://label/email', lazy=False)
            self.assertEqual(node.uri, 'i18n://sv-se@label/email.txt#1')

        cio.delete(self.uri)
        content = cache.get(self.uri)
        self.assertIsNone(content)

    def test_cache_encoding(self):
        cio.set(self.uri, u'epost')

        cached_node = cache.get(self.uri)
        content = cached_node['content']
        self.assertIsInstance(content, unicode)
        self.assertEqual(content, u'epost')

        cache.set('i18n://sv-se@label/email.txt#1', u'epost')
        nodes = cache.get_many((self.uri, self.uri))
        self.assertDictEqual(nodes, {self.uri: {'uri': 'i18n://sv-se@label/email.txt#1', 'content': u'epost'}})

    def test_cache_delete(self):
        uris = ['i18n://sv-se@foo.txt', 'i18n://sv-se@bar.txt']

        cache.set(uris[0], u'Foo')
        cache.set(uris[1], u'Bar')

        with self.assertCache(hits=2):
            cache.get_many(uris)

        cache.delete_many(uris)

        with self.assertCache(misses=2):
            cache.get_many(uris)

    def test_cache_set(self):
        with self.assertRaises(URI.Invalid):
            cache.set('i18n://sv-se@foo', u'Bar')

        nodes = {
            'i18n://sv-se@foo.txt#1': u'Foo',
            'i18n://sv-se@bar.txt#2': u'Bar'
        }
        cache.set_many(nodes)

        with self.assertCache(calls=1, hits=2):
            result = cache.get_many(['i18n://sv-se@foo', 'i18n://sv-se@bar'])
            self.assertDictEqual(result, {
                'i18n://sv-se@foo': {'uri': 'i18n://sv-se@foo.txt#1', 'content': u'Foo'},
                'i18n://sv-se@bar': {'uri': 'i18n://sv-se@bar.txt#2', 'content': u'Bar'}
            })



from time import mktime
from datetime import datetime
from ...pipeline.pipes.base import BasePipe


class MetaPipe(BasePipe):

    def set_request(self, request):
        for node in request.values():
            node.meta['modified_at'] = self._utc_timestamp()

    def publish_request(self, request):
        for node in request.values():
            node.meta['published_at'] = self._utc_timestamp()

    def _utc_timestamp(self):
        return int(mktime(datetime.utcnow().timetuple()))



import cio
from cio.backends import cache
from cio.pipeline import pipeline
from cio.backends import storage
from cio.backends.exceptions import NodeDoesNotExist
from cio.utils.uri import URI
from tests import BaseTest


class ApiTest(BaseTest):

    def test_get(self):
        node = cio.get('label/email', default=u'fallback')
        self.assertEqual(node.content, u'fallback')
        self.assertEqual(node.initial_uri, 'label/email')
        self.assertEqual(node.uri, 'i18n://sv-se@label/email.txt')

    def test_get_with_empty_default(self):
        node = cio.get('page/title', default=u'', lazy=False)
        self.assertEqual(node.content, u'')
        node = cio.get('page/body', default=None, lazy=False)
        self.assertIsNone(node.content)

        # Testing same non existing uri's twice to assert cache handles None/"" default
        node = cio.get('page/title', default=u'', lazy=False)
        self.assertEqual(node.content, u'')
        node = cio.get('page/body', default=None, lazy=False)
        self.assertIsNone(node.content)

    def test_set(self):
        with self.assertRaises(URI.Invalid):
            cio.set('page/title', 'fail')

        with self.assertRaises(URI.Invalid):
            cio.set('page/title.txt', 'fail')

        node = cio.set('i18n://sv-se@label/email.md', u'e-post')
        self.assertEqual(node.uri, 'i18n://sv-se@label/email.md#1')
        cache.clear()
        node = cio.get('label/email', u'fallback')
        self.assertEqual(node.content, u'<p>e-post</p>')
        self.assertEqual(node.uri, 'i18n://sv-se@label/email.md#1')
        self.assertEqual(node.initial, u'fallback')
        self.assertEqual(len(node.meta.keys()), 0)  # No meta returned from non-versioned api get
        self.assertEqual(repr(node._node), '<Node: i18n://sv-se@label/email.md#1>')
        self.assertEqual(node.for_json(), {
            'uri': node.uri,
            'content': node.content,
            'meta': node.meta
        })

        node = cio.set('sv-se@label/email', u'e-post', publish=False)
        self.assertEqual(node.uri, 'i18n://sv-se@label/email.txt#draft')
        self.assertKeys(node.meta, 'modified_at', 'is_published')

        node = cio.publish(node.uri)
        self.assertKeys(node.meta, 'modified_at', 'published_at', 'is_published')
        self.assertTrue(node.meta['is_published'])

        node = cio.get('label/email')
        self.assertEqual(node.uri, 'i18n://sv-se@label/email.txt#2')
        self.assertEqual(node.content, u'e-post')
        self.assertEqual(node.uri.ext, 'txt')
        self.assertEqual(len(node.meta.keys()), 0)

    def test_delete(self):
        with self.assertRaises(URI.Invalid):
            cio.delete('foo/bar')

        node = cio.set('i18n://sv-se@label/email.txt', u'e-post')
        uri = node.uri
        self.assertEqual(cache.get(uri)['content'], u'e-post')

        uris = cio.delete('sv-se@label/email#1', 'sv-se@foo/bar')
        self.assertListEqual(uris, ['sv-se@label/email#1'])

        with self.assertRaises(NodeDoesNotExist):
            storage.get(uri)

        self.assertIsNone(cache.get(uri))

    def test_revisions(self):
        def assertRevisions(*revs):
            revisions = set(cio.revisions('i18n://sv-se@page/title'))
            assert revisions == set(revs)

        self.assertEqual(len(set(cio.revisions('i18n://sv-se@page/title'))), 0)

        node = cio.load('sv-se@page/title')
        self.assertDictEqual(node, {
            'uri': 'i18n://sv-se@page/title.txt',
            'data': None,
            'content': None,
            'meta': {}
        })

        # First draft
        with self.assertDB(selects=1, inserts=1, updates=0):
            with self.assertCache(calls=0):
                node = cio.set('i18n://sv-se@page/title.txt', u'Content-IO', publish=False)
                self.assertEqual(node.uri, 'i18n://sv-se@page/title.txt#draft')

        assertRevisions(('i18n://sv-se@page/title.txt#draft', False))
        self.assertIsNone(cio.get('page/title').content)

        # Publish first draft, version 1
        with self.assertDB(calls=4, selects=2, updates=2):
            with self.assertCache(calls=1, sets=1):
                node = cio.publish(node.uri)
                self.assertEqual(node.uri, 'i18n://sv-se@page/title.txt#1')

        assertRevisions(('i18n://sv-se@page/title.txt#1', True))
        self.assertEqual(cio.get('page/title').content, u'Content-IO')

        # Second draft
        with self.assertDB(selects=1, inserts=1, updates=0):
            with self.assertCache(calls=0):
                node = cio.set('i18n://sv-se@page/title.md', u'# Content-IO - Fast!', publish=False)
                self.assertEqual(node.uri, 'i18n://sv-se@page/title.md#draft')

        assertRevisions(('i18n://sv-se@page/title.txt#1', True), ('i18n://sv-se@page/title.md#draft', False))
        self.assertEqual(cio.get('page/title').content, u'Content-IO')

        # Publish second draft, version 2
        with self.assertDB(calls=4, selects=2, updates=2):
            with self.assertCache(calls=1, sets=1):
                node = cio.publish(node.uri)
                self.assertEqual(node.uri, 'i18n://sv-se@page/title.md#2')

        assertRevisions(('i18n://sv-se@page/title.txt#1', False), ('i18n://sv-se@page/title.md#2', True))
        self.assertEqual(cio.get('page/title').content, u'<h1>Content-IO - Fast!</h1>')

        # Alter published version 2
        with self.assertDB(calls=2, selects=1, inserts=0, updates=1):
            with self.assertCache(calls=0):
                node = cio.set('i18n://sv-se@page/title.md#2', u'# Content-IO - Lightening fast!', publish=False)
                self.assertEqual(node.uri, 'i18n://sv-se@page/title.md#2')

        assertRevisions(('i18n://sv-se@page/title.txt#1', False), ('i18n://sv-se@page/title.md#2', True))
        self.assertEqual(cio.get('page/title').content, u'<h1>Content-IO - Fast!</h1>')  # Not published, still in cache

        # Re-publish version 2, no change
        with self.assertDB(selects=1, inserts=0, updates=0):
            with self.assertCache(calls=1, sets=1):
                node = cio.publish(node.uri)
                self.assertEqual(node.uri, 'i18n://sv-se@page/title.md#2')

        assertRevisions(('i18n://sv-se@page/title.txt#1', False), ('i18n://sv-se@page/title.md#2', True))
        self.assertEqual(cio.get('page/title').content, u'<h1>Content-IO - Lightening fast!</h1>')

        # Rollback version 1
        with self.assertDB(calls=3, selects=1, updates=2):
            with self.assertCache(calls=1, sets=1):
                node = cio.publish('i18n://sv-se@page/title#1')
                self.assertEqual(node.uri, 'i18n://sv-se@page/title.txt#1')

        assertRevisions(('i18n://sv-se@page/title.txt#1', True), ('i18n://sv-se@page/title.md#2', False))
        self.assertEqual(cio.get('page/title').content, u'Content-IO')

        # Assert get specific version doesn't mess up the cache
        cache.clear()
        with self.assertCache(calls=0):
            self.assertEqual(cio.get('page/title#2').content, u'<h1>Content-IO - Lightening fast!</h1>')
        with self.assertCache(calls=2, misses=1, sets=1):
            self.assertEqual(cio.get('page/title').content, u'Content-IO')

        # Load version 1 and 2
        data = cio.load('sv-se@page/title#1')
        self.assertEqual(data['uri'], 'i18n://sv-se@page/title.txt#1')
        self.assertEqual(data['data'], u'Content-IO')
        data = cio.load('sv-se@page/title#2')
        self.assertEqual(data['uri'], 'i18n://sv-se@page/title.md#2')
        self.assertEqual(data['data'], u'# Content-IO - Lightening fast!')

        # Load without version and expect published version
        data = cio.load('sv-se@page/title')
        self.assertEqual(data['uri'], 'i18n://sv-se@page/title.txt#1')
        self.assertEqual(data['data'], u'Content-IO')

    def test_environment_state(self):
        with cio.env(i18n='en-us'):
            node = cio.get('page/title')
            self.assertEqual(node.uri, 'i18n://en-us@page/title.txt')

        node = cio.get('page/title')
        self.assertEqual(node.uri, 'i18n://sv-se@page/title.txt')

    def test_non_distinct_uri(self):
        node1 = cio.get('page/title', u'Title1')
        node2 = cio.get('page/title', u'Title2')
        self.assertEqual(unicode(node1), u'Title1')
        self.assertEqual(unicode(node2), u'Title1')

        node1 = cio.get('page/title', u'Title1', lazy=False)
        cache.clear()
        node2 = cio.get('page/title', u'Title2', lazy=False)
        self.assertEqual(unicode(node1), u'Title1')
        self.assertEqual(unicode(node2), u'Title2')  # Second node not buffered, therefore unique default content

    def test_fallback(self):
        with cio.env(i18n=('sv-se', 'en-us', 'en-uk')):
            node1 = cio.set('i18n://bogus@label/email.txt', u'epost')
            node2 = cio.set('i18n://en-uk@label/surname.txt', u'surname')

            with self.assertCache(misses=2, sets=2):
                with self.assertDB(calls=6, selects=6):
                    node3 = cio.get('i18n://label/email')
                    node4 = cio.get('i18n://label/surname', u'efternamn')
                    self.assertIsNone(node3.content)
                    self.assertEqual(node4.content, u'surname')

            cache.clear()

            with self.assertCache(misses=2, sets=2):
                with self.assertDB(calls=6):
                    cio.get('i18n://label/email', lazy=False)
                    cio.get('i18n://label/surname', u'lastname', lazy=False)

    def test_uri_redirect(self):
        cio.set('i18n://sv-se@page/title.txt', u'Title')

        node = cio.get('i18n://sv-se@page/title', u'Default')
        self.assertEqual(node.uri, 'i18n://sv-se@page/title.txt#1')
        self.assertEqual(node.content, u'Title')

        node = cio.get('i18n://sv-se@page/title.md', u'# Default Markdown', lazy=False)
        self.assertEqual(node.uri, 'i18n://sv-se@page/title.md')
        self.assertEqual(node.content, u'<h1>Default Markdown</h1>')  # Cache still contains 'Title', but plugin diff and skipped
        cached_node = cache.get(node.uri)
        self.assertDictEqual(cached_node, {'uri': node.uri, 'content': u'<h1>Default Markdown</h1>'})

        cache.clear()
        node = cio.get('i18n://sv-se@page/title.md', u'Default Markdown', lazy=False)
        self.assertEqual(node.uri, 'i18n://sv-se@page/title.md')
        self.assertEqual(node.content, u'<p>Default Markdown</p>')  # Cache cleared, storage plugin mismatch, default fallback

    def test_node_meta(self):
        node = cio.set('sv-se@page/title', u'', author=u'lundberg')
        self.assertEqual(node.meta.get('author'), u'lundberg')

        node = cio.get('page/title')
        self.assertEqual(len(node.meta.keys()), 0)  # Cached node has no meta

        node = cio.load('sv-se@page/title#1')
        meta = node['meta']
        self.assertKeys(meta, 'author', 'modified_at', 'published_at', 'is_published')
        self.assertEqual(meta.get('author'), u'lundberg')

        cio.set('sv-se@page/title#1', u'', comment=u'This works!')
        node = cio.load('sv-se@page/title#1')
        meta = node['meta']
        self.assertKeys(meta, 'author', 'comment', 'modified_at', 'published_at', 'is_published')
        self.assertEqual(meta.get('author'), u'lundberg')
        self.assertEqual(meta.get('comment'), u'This works!')

        cio.set('sv-se@page/title#1', u'', comment=None)
        node = cio.load('sv-se@page/title#1')
        meta = node['meta']
        self.assertKeys(meta, 'author', 'modified_at', 'published_at', 'is_published')
        self.assertEqual(meta.get('author'), u'lundberg')
        self.assertNotIn('comment', meta)

    def test_pipes_hits(self):
        with cio.env(i18n=('sv-se', 'en-us')):
            with self.assertDB(inserts=2):
                with self.assertCache(calls=2, sets=2):
                    cio.set('i18n://sv-se@label/email.txt', u'epost')
                    cio.set('i18n://en-us@label/surname.txt', u'surname')

            # Lazy gets
            with self.assertDB(calls=0):
                with self.assertCache(calls=0):
                    node1 = cio.get('label/email')
                    node2 = cio.get('i18n://label/surname')
                    node3 = cio.get('i18n://monkey@label/zipcode', default=u'postnummer')

            # with self.assertDB(calls=2), self.assertCache(calls=5, hits=1, misses=2, sets=2):
            with self.assertDB(calls=4, selects=4):
                with self.assertCache(calls=2, hits=1, misses=2, sets=2):
                    self.assertEqual(unicode(node1), u'epost')
                    self.assertEqual(node2.content, u'surname')
                    self.assertEqual(unicode(node3), u'postnummer')

            with self.assertDB(calls=0):
                with self.assertCache(calls=1, hits=3):
                    node1 = cio.get('label/email')
                    node2 = cio.get('i18n://label/surname')
                    node3 = cio.get('i18n://monkey@label/zipcode', default=u'postnummer')
                    self.assertEqual(unicode(node1), u'epost')
                    self.assertEqual(node2.content, u'surname')
                    self.assertEqual(unicode(node3), u'postnummer')

            self.assertIsNotNone(repr(node1))
            self.assertIsNotNone(str(node1))

    def test_forced_empty_content(self):
        with self.assertRaises(ValueError):
            cio.set('i18n://sv-se@none', None)

        node = cio.set('i18n://sv-se@empty.txt', u'')
        node = cio.get(node.uri, default=u'fallback')
        self.assertEqual(unicode(node), u'')

    def test_load_pipeline(self):
        with self.assertRaises(ImportError):
            pipeline.add_pipe('foo.Bar')



import sys


def import_module(package):
    __import__(package)
    return sys.modules[package]


def import_class(import_path, name=None):
    """
    Imports and returns class for full class path string.
    Ex. 'foo.bar.Bogus' -> <class 'foo.bar.Bogus'>
    """
    if not name:
        import_path, name = import_path.rsplit('.', 1)
    mod = import_module(import_path)
    try:
        return getattr(mod, name)
    except AttributeError as e:
        raise ImportError(e)



# coding=utf-8
import sys

VERSION = (1, 0, 1, 'final', 0)


def get_version(version=None):
    """Derives a PEP386-compliant version number from VERSION."""
    if version is None:
        version = VERSION
    assert len(version) == 5
    assert version[3] in ('alpha', 'beta', 'rc', 'final')

    # Now build the two parts of the version number:
    # main = X.Y[.Z]
    # sub = .devN - for pre-alpha releases
    #     | {a|b|c}N - for alpha, beta and rc releases

    parts = 2 if version[2] == 0 else 3
    main = '.'.join(str(x) for x in version[:parts])

    sub = ''
    if version[3] != 'final':
        mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'c'}
        sub = mapping[version[3]] + str(version[4])

    return main + sub

__version__ = get_version()


class lazy_shortcut(object):
    def __init__(self, module, target):
        self._evaluated = False
        self.module = module
        self.targets = target.split('.', 1)

    def __call__(self, *args, **kwargs):
        target = self.__evaluate__()
        return target(*args, **kwargs)

    def __getattr__(self, item):
        target = self.__evaluate__()
        return getattr(target, item)

    def __evaluate__(self):
        if not self._evaluated:
            from .utils.imports import import_module

            # Evaluate target
            target = import_module(self.module)
            for name in self.targets:
                target = getattr(target, name)

            # Point module shortcut variable to target
            __module__ = sys.modules[__name__]
            setattr(__module__, name, target)

            self._evaluated = True
            return target


env = lazy_shortcut('cio.environment', 'env')
get = lazy_shortcut('cio.api', 'get')
set = lazy_shortcut('cio.api', 'set')
load = lazy_shortcut('cio.api', 'load')
delete = lazy_shortcut('cio.api', 'delete')
publish = lazy_shortcut('cio.api', 'publish')
revisions = lazy_shortcut('cio.api', 'revisions')



import logging
from contextlib import contextmanager
from types import ModuleType
from . import default_settings

logger = logging.getLogger(__name__)


class Settings(dict):

    def __init__(self, conf=None, **settings):
        self._listeners = set()
        self.configure(conf=conf, **settings)

    @contextmanager
    def __call__(self, **settings):
        state = self.deepcopy()
        self.configure(**settings)
        yield
        self.clear()
        self.update(state)

    def deepcopy(self):
        copy = {}
        for key, value in self.iteritems():
            if isinstance(value, dict):
                value = dict(value)
            if isinstance(value, list):
                value = list(value)
            copy[key] = value
        return copy

    def configure(self, conf=None, **settings):
        if isinstance(conf, ModuleType):
            conf = conf.__dict__

        for setting, value in (conf or settings).iteritems():
            if setting.isupper():
                self[setting] = value

        for callback in self._listeners:
            try:
                callback()
            except Exception as e:
                logger.warn('Failed to notify callback about new settings; %s', e)

    def watch(self, callback):
        self._listeners.add(callback)

    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


settings = Settings(default_settings)



from .conf import settings
from .environment import env
from .node import Node, empty
from .pipeline import pipeline
from .plugins import plugins
from .backends import storage
from .backends.exceptions import NodeDoesNotExist
from .utils.uri import URI

__all__ = ['get', 'set', 'delete', 'publish', 'revisions', 'load']


def get(uri, default=None, lazy=True):

    node = Node(uri, default)

    # Set URI namespace to current environment scheme, if not set
    uri = node.uri
    if not uri.namespace:
        namespace = getattr(env, node.uri.scheme)[0]
        uri = uri.clone(namespace=namespace)
    node.uri = uri

    # Send/Buffer node through pipeline
    if lazy:
        node = pipeline.buffer('get', node)
    else:
        # TODO: Return node from pipeline
        pipeline.send('get', node)

    return node


def set(uri, data, publish=True, **meta):
    node = Node(uri, data, **meta)

    # Extend uri with missing extension and version
    uri = node.uri
    if not uri.ext:
        uri = uri.clone(ext=settings.URI_DEFAULT_EXT)
    if not uri.version:
        uri = uri.clone(version='draft')
    node.uri = uri

    # Send node through pipeline
    pipeline.send('set', node)

    # Auto publish
    if publish:
        pipeline.send('publish', node)

    return node


def delete(*uris):
    # Initialize nodes with "empty" content
    def init_node(uri):
        node = Node(uri, empty)

        # Default version to draft
        if not node.uri.version:
            node.uri = node.uri.clone(version='draft')
        return node
    nodes = (init_node(uri) for uri in uris)

    # Send nodes through pipeline
    response = pipeline.send('delete', *nodes)

    # Return requested uris for successfully deleted nodes (content set to None)
    return [node.initial_uri for node in response.values() if node.content is None]


def publish(uri):
    node = Node(uri)

    # Publish draft if no specific version specified
    if not node.uri.version:
        node.uri = node.uri.clone(version='draft')

    response = pipeline.send('publish', node)
    return response.get(uri)


def revisions(uri):
    return storage.get_revisions(uri)


def load(uri):
    uri = URI(uri)
    node = None

    def uri_chain(uri):
        if uri.version:
            yield uri
        if uri.version != 'draft':
            yield uri.clone(version='draft')
        yield uri.clone(version=None)

    # Try to get node from storage in order: given version, draft, published
    for _uri in uri_chain(uri):
        try:
            node = storage.get(_uri)
        except NodeDoesNotExist:
            continue
        else:
            break

    if node:
        # Load node data with related plugin
        source = node.pop('content')
        plugin = plugins.resolve(node['uri'])  # May raise UnknownPlugin and should be handled outside api
        data = node['data'] = plugin.load(source)
        node['content'] = plugin.render(data)

    else:
        # Initialize non-existing node without version
        uri = uri.clone(version=None)

        # Set default extension
        if not uri.ext:
            uri = uri.clone(ext=settings.URI_DEFAULT_EXT)

        # Validate plugin existence
        plugins.resolve(uri)

        node = {
            'uri': uri,
            'data': None,
            'content': None,
            'meta': {}
        }

    return node



from ..base import CacheBackend


class LocMemCacheBackend(CacheBackend):

    scheme = 'locmem'

    def __init__(self, **config):
        super(LocMemCacheBackend, self).__init__(**config)
        self._cache = {}
        self.calls = 0
        self.hits = 0
        self.misses = 0
        self.sets = 0

    def clear(self):
        self._cache.clear()

    def _get(self, key):
        value = self._cache.get(key)
        self.calls += 1
        if value is None:
            self.misses += 1
        else:
            self.hits += 1
        return value

    def _get_many(self, keys):
        result = {}
        for key in keys:
            value = self._cache.get(key)
            if value is not None:
                result[key] = value
                self.hits += 1
            else:
                self.misses += 1
        self.calls += 1
        return result

    def _set(self, key, value):
        self._cache[key] = value
        self.calls += 1
        self.sets += 1

    def _set_many(self, data):
        for key, value in data.iteritems():
            self._cache[key] = value
            self.sets += 1
        self.calls += 1

    def _delete(self, key):
        if key in self._cache:
            del self._cache[key]
        self.calls += 1

    def _delete_many(self, keys):
        for key in keys:
            if key in self._cache:
                del self._cache[key]
        self.calls += 1



class BasePipe(object):
    """
    Optional implementable pipe methods:

    def get_request(self, request):
        pass

    def get_response(self, response):
        return response

    def set_request(self, request):
        pass

    def set_response(self, response):
        return response

    def delete_request(self, request):
        pass

    def delete_response(self, response):
        return response

    def publish_request(self, request):
        pass

    def publish_response(self, response):
        return response
    """

    def materialize_node(self, node, uri, content, meta=None):
        """
        Set node uri and content from backend
        """
        node.uri = uri
        node.content = content
        node.meta = meta if meta is not None else {}



import inspect
from .base import BaseBackend, CacheBackend, StorageBackend
from .exceptions import InvalidBackend
from ..conf import settings
from ..utils.imports import import_class
from ..utils.uri import URI

BACKENDS = {
    'locmem': 'locmem',
    'sqlite': 'sqlite'
}


def get_backend(backend):
    config = {}

    # Unpack backend dict format
    if isinstance(backend, dict):
        _backend = backend.get('BACKEND')
        config.update(backend)
        backend = _backend

    if inspect.isclass(backend) and issubclass(backend, BaseBackend):
        backend_class = backend
    else:
        # Parse uri or package
        if '://' in backend:
            scheme, _config = backend.split('://', 1)
            if scheme not in BACKENDS:
                raise InvalidBackend('Invalid content-io backend scheme "%s"' % scheme)
            package = backend = 'cio.backends.%s' % BACKENDS[scheme]
            class_name = 'Backend'

            # Parse config
            name, _, params = _config.partition('?')
            config['NAME'] = name
            if params:
                config.update(dict(param.split('=') for param in params.split('&')))
        elif '.' in backend:
            package, class_name = backend.rsplit('.', 1)
        else:
            raise InvalidBackend('Invalid content-io backend "%s"' % backend)

        # Import and instantiate backend
        try:
            backend_class = import_class(package, class_name)
        except ImportError as e:
            raise ImportError('Could not import content-io backend "%s" (Is it on sys.path?): %s' % (backend, e))

    return backend_class(**config)


class BackendManager(object):
    """
    Manager for backend. Handles arg validation.
    """
    def __init__(self):
        self._backend = None
        settings.watch(self.setup)

    @property
    def backend(self):
        if not self._backend:
            self.setup()

        return self._backend

    def setup(self):
        # Find and instantiate backend
        config = self._get_backend_config()
        backend = get_backend(config)

        # Validate backend
        if self._is_valid_backend(backend):
            self._backend = backend
        else:
            raise InvalidBackend('Invalid content-io %s backend "%s"' % (self._scope(), self._conf))

    def _scope(self):
        return self.__class__.__name__.rstrip('Manager').lower()

    def _get_backend_config(self):
        raise NotImplementedError

    def _is_valid_backend(self, backend):
        raise NotImplementedError

    def _clean_get_uri(self, uri):
        raise NotImplementedError

    def _clean_set_uri(self, uri):
        raise NotImplementedError

    def _clean_delete_uri(self, uri):
        raise NotImplementedError

    def _clean_get_uris(self, uris):
        return tuple(self._clean_get_uri(uri) for uri in uris)

    def _clean_set_uris(self, uris):
        return tuple(self._clean_set_uri(uri) for uri in uris)

    def _clean_delete_uris(self, uris):
        return tuple(self._clean_delete_uri(uri) for uri in uris)

    def _clean_uri(self, uri, *parts):
        uri = URI(uri)
        if not uri.has_parts(*parts):
            raise URI.Invalid('Invalid URI "%s"; must contain %s.' % (uri, ', '.join(parts)))
        return uri


class CacheManager(BackendManager, CacheBackend):

    def _get_backend_config(self):
        return settings.CACHE

    def get(self, uri):
        uri = self._clean_get_uri(uri)
        return self.backend.get(uri)

    def get_many(self, uris):
        uris = self._clean_get_uris(uris)
        return self.backend.get_many(uris)

    def set(self, uri, content):
        uri = self._clean_set_uri(uri)
        self.backend.set(uri, content)

    def set_many(self, nodes):
        nodes = dict((self._clean_set_uri(uri), content) for uri, content in nodes.iteritems())
        self.backend.set_many(nodes)

    def delete(self, uri):
        uri = self._clean_delete_uri(uri)
        self.backend.delete(uri)

    def delete_many(self, uris):
        uris = self._clean_delete_uris(uris)
        self.backend.delete_many(uris)

    def clear(self):
        self.backend.clear()

    def _is_valid_backend(self, backend):
        return isinstance(backend, CacheBackend)

    def _clean_get_uri(self, uri):
        return self._clean_uri(uri, 'namespace', 'path')

    def _clean_set_uri(self, uri):
        return self._clean_uri(uri, 'namespace', 'path', 'ext')

    def _clean_delete_uri(self, uri):
        return self._clean_uri(uri, 'namespace', 'path')


class StorageManager(BackendManager, StorageBackend):

    def _get_backend_config(self):
        return settings.STORAGE

    def get(self, uri):
        uri = self._clean_get_uri(uri)
        return self.backend.get(uri)

    def get_many(self, uris):
        uris = self._clean_get_uris(uris)
        return self.backend.get_many(uris)

    def set(self, uri, content, **meta):
        uri = self._clean_set_uri(uri)

        if content is None:
            raise ValueError('Can not persist content equal to None for URI "%s".' % uri)

        return self.backend.set(uri, content, **meta)

    def delete(self, uri):
        uri = self._clean_delete_uri(uri)
        return self.backend.delete(uri)

    def delete_many(self, uris):
        uris = self._clean_delete_uris(uris)
        return self.backend.delete_many(uris)

    def publish(self, uri, **meta):
        uri = self._clean_publish_uri(uri)
        return self.backend.publish(uri, **meta)

    def get_revisions(self, uri):
        uri = self._clean_get_uri(uri)
        return self.backend.get_revisions(uri)

    def _is_valid_backend(self, backend):
        return isinstance(backend, StorageBackend)

    def _clean_get_uri(self, uri):
        return self._clean_uri(uri, 'namespace', 'path')

    def _clean_set_uri(self, uri):
        return self._clean_uri(uri, 'namespace', 'path', 'ext', 'version')

    def _clean_delete_uri(self, uri):
        return self._clean_uri(uri, 'namespace', 'path', 'version')

    def _clean_publish_uri(self, uri):
        return self._clean_uri(uri, 'namespace', 'path', 'version')


cache = CacheManager()
storage = StorageManager()



from .exceptions import UnknownPlugin
from ..conf import settings
from ..utils.imports import import_class
from ..utils.uri import URI


class PluginLibrary(object):

    def __init__(self):
        self._plugins = {}
        settings.watch(self.load)

    def __iter__(self):
        return self.plugins.iterkeys()

    @property
    def plugins(self):
        if not self._plugins:
            self.load()
        return self._plugins

    def load(self):
        self._plugins = {}
        for plugin_path in settings.PLUGINS:
            self.register(plugin_path)

    def register(self, plugin):
        if isinstance(plugin, basestring):
            try:
                plugin_class = import_class(plugin)
                self.register(plugin_class)
            except ImportError as e:
                raise ImportError('Could not import content-io plugin "%s" (Is it on sys.path?): %s' % (plugin, e))
        else:
            self._plugins[plugin.ext] = plugin()

    def get(self, ext):
        if ext not in self.plugins:
            raise UnknownPlugin
        return self.plugins[ext]

    def resolve(self, uri):
        uri = URI(uri)
        return self.get(uri.ext)


plugins = PluginLibrary()



from .base import BasePipe
from ...conf.exceptions import ImproperlyConfigured
from ...node import empty
from ...plugins import plugins
from ...plugins.exceptions import UnknownPlugin


class PluginPipe(BasePipe):

    def render_response(self, response):
        for node in response.values():
            try:
                plugin = plugins.resolve(node.uri)
            except UnknownPlugin:
                raise ImproperlyConfigured('Unknown plugin "%s" or improperly configured pipeline for node "%s".' % (
                    node.uri.ext,
                    node.uri
                ))
            else:
                data = plugin.load(node.content)
                node.content = plugin.render(data)

        return response

    def get_response(self, response):
        return self.render_response(response)

    def set_request(self, request):
        for node in request.values():
            try:
                plugin = plugins.resolve(node.uri)
            except UnknownPlugin:
                pass
                # TODO: Should we maybe raise here?
            else:
                node.content = plugin.save(node.content)

    def set_response(self, response):
        return self.render_response(response)

    def publish_response(self, response):
        return self.render_response(response)

    def delete_response(self, response):
        for node in response.values():
            try:
                plugin = plugins.resolve(node.uri)
                if node.content is not empty:
                    plugin.delete(node.content)
            except UnknownPlugin:
                pass
                # TODO: Should we maybe raise here?

        return response



ENVIRONMENT = {
    'default': {
        'i18n': 'en-us',
        'l10n': 'local',
        'g11n': 'global'
    }
}

CACHE = 'locmem://'
STORAGE = 'sqlite://:memory:'

PIPELINE = [
    'cio.pipeline.pipes.cache.CachePipe',
    'cio.pipeline.pipes.meta.MetaPipe',
    'cio.pipeline.pipes.plugin.PluginPipe',
    'cio.pipeline.pipes.storage.StoragePipe',
    'cio.pipeline.pipes.storage.NamespaceFallbackPipe'
]

PLUGINS = [
    'cio.plugins.txt.TextPlugin',
    'cio.plugins.md.MarkdownPlugin'
]

URI_SCHEME_SEPARATOR = '://'
URI_NAMESPACE_SEPARATOR = '@'
URI_PATH_SEPARATOR = '/'
URI_EXT_SEPARATOR = '.'
URI_VERSION_SEPARATOR = '#'
URI_DEFAULT_EXT = 'txt'



import threading
import cio
from cio import get_version
from cio.conf import settings, default_settings
from cio.pipeline import pipeline
from tests import BaseTest


class AppTest(BaseTest):

    def test_version(self):
        self.assertEqual(get_version((1, 2, 3, 'alpha', 1)), '1.2.3a1')
        self.assertEqual(get_version((1, 2, 3, 'beta', 2)), '1.2.3b2')
        self.assertEqual(get_version((1, 2, 3, 'rc', 3)), '1.2.3c3')
        self.assertEqual(get_version((1, 2, 3, 'final', 4)), '1.2.3')

    def test_settings(self):
        self.assertEqual(settings.URI_NAMESPACE_SEPARATOR, default_settings.URI_NAMESPACE_SEPARATOR)
        pre = settings.STORAGE
        with settings():
            settings.configure(STORAGE='bogus.newstorage')
            self.assertEqual(settings.STORAGE, 'bogus.newstorage')
        self.assertEqual(settings.STORAGE, pre)

    def test_environment(self):
        """
        'default': {
            'g11n': 'global',
            'i18n': (django_settings.LANGUAGE_CODE,),
            'l10n': (project_name(django_settings) or 'local',)
        }
        """
        env = settings.ENVIRONMENT
        self.assertDictEqual(env['default'], {'g11n': 'global', 'i18n': 'sv-se', 'l10n': 'tests'})

        with self.assertRaises(IndexError):
            cio.env.pop()

        with cio.env(i18n='sv', l10n=['loc'], g11n=('glob',)):
            self.assertTupleEqual(cio.env.i18n, ('sv',))
            self.assertTupleEqual(cio.env.l10n, ('loc',))
            self.assertTupleEqual(cio.env.g11n, ('glob',))

        with self.assertRaises(SystemError):
            cio.env.__init__()

    def test_context(self):
        self.assertTupleEqual(cio.env.state.i18n, ('sv-se',))

        with settings():
            settings.ENVIRONMENT['bogus'] = {
                'g11n': 'global',
                'i18n': ('sv', 'en'),
                'l10n': ('foo', 'bar')
            }

            with cio.env('bogus'):
                self.assertTupleEqual(cio.env.state.i18n, ('sv', 'en'))
                self.assertTupleEqual(cio.env.state.l10n, ('foo', 'bar'))

            with cio.env(i18n='en-us'):
                node = cio.get('i18n://label/firstname', lazy=False)
                buffered_node = cio.get('i18n://label/email')

                self.assertTupleEqual(cio.env.i18n, ('en-us',))
                self.assertEqual(len(pipeline._buffer), 1)
                self.assertEqual(len(pipeline.history), 1)

                def assert_new_thread_env():
                    self.assertTupleEqual(cio.env.i18n, ('sv-se',))
                    self.assertEqual(len(pipeline._buffer), 0)
                    self.assertEqual(len(pipeline.history), 0)
                    cio.get('i18n://label/surname', lazy=False)
                    self.assertEqual(len(pipeline.history), 1)

                thread = threading.Thread(target=assert_new_thread_env)
                thread.start()
                thread.join()

                buffered_node.flush()
                self.assertEqual(len(pipeline._buffer), 0)
                self.assertEqual(len(pipeline.history), 2)
                self.assertListEqual(pipeline.history.list('get'), [node, buffered_node._node])

        self.assertNotIn('bogus', settings.ENVIRONMENT.keys())



import sqlite3
from itertools import izip
from sqlite3 import IntegrityError
from ..exceptions import NodeDoesNotExist, PersistenceError
from ...backends.base import DatabaseBackend
from ...conf.exceptions import ImproperlyConfigured


class SqliteBackend(DatabaseBackend):

    def __init__(self, **config):
        super(SqliteBackend, self).__init__(**config)
        self.debug = False
        self.queries = []
        if 'NAME' not in self.config:
            raise ImproperlyConfigured('Missing sqlite database name.')
        database = self.config['NAME']
        kwargs = self.config.get('OPTIONS', {})
        self._connection = sqlite3.connect(database, **kwargs)
        self._setup()

    def _setup(self):
        with self._connection as con:
            con.execute("""
                CREATE TABLE IF NOT EXISTS "content_io_node" (
                    "id" integer NOT NULL PRIMARY KEY ASC AUTOINCREMENT,
                    "key" varchar(255) NOT NULL,
                    "content" text NOT NULL,
                    "plugin" varchar(8) NOT NULL,
                    "version" varchar(255) NOT NULL,
                    "is_published" bool NOT NULL,
                    "meta" text
                );
            """)
            con.execute('CREATE INDEX IF NOT EXISTS "content_io_node_key" ON "content_io_node" ("key");')

    def start_debug(self):
        self.queries = []
        self.debug = True

    def stop_debug(self):
        self.queries = []
        self.debug = False

    def _call(self, command, query, **params):
        with self._connection as con:
            cursor = con.cursor()
            sql = command + ' ' + query
            cursor.execute(sql, params)
            if self.debug:
                self.queries.append({'sql': sql, 'params': params})
            return cursor

    def _call_select(self, query, **params):
        return self._call('SELECT', query, **params)

    def _call_insert(self, query, **params):
        self._call('INSERT INTO content_io_node', query, **params)

    def _call_update(self, query, **params):
        self._call('UPDATE content_io_node SET', query, **params)

    def _call_delete(self, where='', **params):
        command = 'DELETE FROM content_io_node'
        if where:
            command += ' WHERE'
        self._call(command, where, **params)

    def publish(self, uri, **meta):
        node = self._get(uri)

        if not node['is_published']:
            # Assign version number
            if not node['version'].isdigit():
                result = self._call_select('version FROM content_io_node WHERE key=:key', key=node['key'])
                revisions = (r[0] for r in result.fetchall())
                version = self._get_next_version(revisions)
                node['version'] = version

            # Un publish any other revision
            self._call_update('is_published=0 WHERE key=:key', key=node['key'])

            # Publish this version
            node['meta'] = self._merge_meta(node['meta'], meta)
            node['is_published'] = 1
            self._call_update('version=:version, is_published=1, meta=:meta WHERE id=:id',
                              id=node['id'],
                              version=node['version'],
                              meta=node['meta'])

        return self._serialize(uri, node)

    def get_revisions(self, uri):
        key = self._build_key(uri)
        nodes = self._call_select('plugin, version, is_published FROM content_io_node WHERE key=:key', key=key)
        return [(uri.clone(ext=ext, version=ver), bool(pub)) for ext, ver, pub in nodes.fetchall()]

    def _get(self, uri):
        columns = ('id', 'key', 'content', 'plugin', 'version', 'is_published', 'meta')
        query = ', '.join(columns) + ' FROM content_io_node WHERE '
        statements = ['key=:key']
        params = {'key': self._build_key(uri)}

        if uri.ext:
            statements.append('plugin=:plugin')
            params['plugin'] = uri.ext
        if uri.version:
            statements.append('version=:version')
            params['version'] = uri.version
        else:
            statements.append('is_published=1')

        query += ' AND '.join(statements)
        result = self._call_select(query, **params)
        node = result.fetchone()

        if node is None:
            raise NodeDoesNotExist('Node for uri "%s" does not exist' % uri)
        else:
            return dict((c, v) for c, v in izip(columns, node))

    def _create(self, uri, content, **meta):
        node = {
            'key': self._build_key(uri),
            'content': content,
            'plugin': uri.ext,
            'version': uri.version,
            'is_published': 0,
            'meta': self._encode_meta(meta)
        }
        try:
            self._call_insert("""
                              (key, content, plugin, version, is_published, meta) VALUES
                              (:key, :content, :plugin, :version, 0, :meta)
                              """, **node)
        except IntegrityError as e:
            raise PersistenceError('Failed to create node for uri "%s"; %s' % (uri, e))

        return node

    def _update(self, uri, content, **meta):
        node = self._get(uri)
        node.update({
            'content': content,
            'plugin': uri.ext,
            'version': uri.version,
            'meta': self._merge_meta(node['meta'], meta)
        })
        self._call_update("""
                          content=:content, plugin=:plugin, version=:version, meta=:meta
                          WHERE id=:id
                          """, **node)
        return node

    def _delete(self, node):
        self._call_delete('id=:id', id=node['id'])



#!/usr/bin/env python
from setuptools import setup, find_packages


tests_require = [
    'unittest2',
    'coverage',
    'Markdown'
]

version = __import__('cio').__version__

setup(
    name='content-io',
    version=version,
    description='Send content through a highly configurable pipeline including cache, plugin and storage pipes',
    long_description=(
        '.. image:: https://travis-ci.org/5monkeys/content-io.png?branch=master\n'
        '    :target: https://travis-ci.org/5monkeys/content-io\n'
        '.. image:: https://coveralls.io/repos/5monkeys/content-io/badge.png?branch=master\n'
        '    :target: https://coveralls.io/r/5monkeys/content-io?branch=master\n\n'
    ),
    author='Jonas Lundberg',
    author_email='jonas@5monkeys.se',
    url='https://github.com/5monkeys/content-io',
    download_url='https://github.com/5monkeys/content-io/tarball/%s' % version,
    keywords=['cms', 'content', 'management', 'pipeline', 'plugin', 'backend', 'cache', 'storage'],
    license='BSD',
    packages=find_packages(exclude='tests'),
    include_package_data=False,
    zip_safe=False,
    classifiers=[
        'Environment :: Web Environment',
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    extras_require={
        'tests': tests_require,
    },
    tests_require=tests_require,
    test_suite='runtests.main',
)



import json
import logging
from hashlib import sha1
from .exceptions import NodeDoesNotExist
from ..utils.uri import URI

logger = logging.getLogger(__name__)


class BaseBackend(object):

    scheme = None

    def __init__(self, **config):
        self.config = config


class CacheBackend(BaseBackend):

    NONE = '__None__'

    def get(self, uri):
        """
        Return node for uri or None if not exists:
            {uri: x, content: y}
        """
        cache_key = self._build_cache_key(uri)
        value = self._get(cache_key)
        if value is not None:
            return self._decode_node(uri, value)

    def get_many(self, uris):
        """
        Return request uri map of found nodes as dicts:
            {requested_uri: {uri: x, content: y}}
        """
        cache_keys = dict((self._build_cache_key(uri), uri) for uri in uris)
        result = self._get_many(cache_keys)
        nodes = {}
        for cache_key in result:
            uri = cache_keys[cache_key]
            value = result[cache_key]
            node = self._decode_node(uri, value)
            if node:
                nodes[uri] = node
        return nodes

    def set(self, uri, content):
        """
        Cache node content for uri.
        No return.
        """
        key, value = self._prepare_node(uri, content)
        self._set(key, value)

    def set_many(self, nodes):
        """
        Takes nodes dict {uri: content, ...} as argument.
        No return.
        """
        data = self._prepare_nodes(nodes)
        self._set_many(data)

    def delete(self, uri):
        """
        Remove node uri from cache.
        No return.
        """
        cache_key = self._build_cache_key(uri)
        self._delete(cache_key)

    def delete_many(self, uris):
        """
        Remove many nodes from cache.
        No return.
        """
        cache_keys = (self._build_cache_key(uri) for uri in uris)
        self._delete_many(cache_keys)

    def clear(self):
        """
        Removes all nodes from cache
        """
        raise NotImplementedError

    def _build_cache_key(self, uri):
        """
        Build sha1 hex cache key to handle key length and whitespace to be compatible with Memcached
        """
        key = uri.clone(ext=None, version=None)
        return sha1(key).hexdigest()

    def _get(self, key):
        raise NotImplementedError

    def _get_many(self, keys):
        raise NotImplementedError

    def _set(self, key, value):
        raise NotImplementedError

    def _set_many(self, data):
        raise NotImplementedError

    def _delete(self, key):
        raise NotImplementedError

    def _delete_many(self, keys):
        raise NotImplementedError

    def _encode_content(self, uri, content):
        """
        Encode/pack node uri and content in a way that the cache backend are able to persist.
        """
        return uri, content

    def _decode_content(self, content):
        """
        Decode/unpack cached node to uri and unicode content.
        """
        uri, content = content
        return uri, content

    def _decode_node(self, uri, content):
        _uri, _content = self._decode_content(content)
        if _uri:
            _uri = URI(_uri)
            if uri.ext in (None, _uri.ext):  # Validate plugin
                return {
                    'uri': _uri,
                    'content': _content
                }

    def _prepare_node(self, uri, content):
        key = self._build_cache_key(uri)
        value = self._encode_content(uri, content)
        return key, value

    def _prepare_nodes(self, nodes):
        return dict(self._prepare_node(uri, content) for uri, content in nodes.iteritems())


class StorageBackend(BaseBackend):

    def get(self, uri):
        """
        Return node for uri or raise NodeDoesNotExist:
            {uri: x, content: y, meta: {}}
        """
        raise NotImplementedError

    def get_many(self, uris):
        """
        Return request uri map of found nodes as dicts:
            {requested_uri: {uri: x, content: y, meta: {}}}
        """
        raise NotImplementedError

    def set(self, uri, content, **meta):
        """
        Persist node data and meta for uri.
        Return tuple of node dict and True if created and False if updated:
            {uri: x, content: y, meta: {}}, True|False
        """
        raise NotImplementedError

    def delete(self, uri):
        """
        Delete node for uri and return node dict or None if not exists:
            {uri: x, content: y, meta: {}}
        """
        raise NotImplementedError

    def delete_many(self, uris):
        """
        Delete node for uri and return request uri map of deleted nodes as dicts:
            {requested_uri: {uri: x, content: y, meta: {}}}
        """
        raise NotImplementedError

    def publish(self, uri, **meta):
        """
        Return published node as dict or raise NodeDoesNotExist:
            {uri: x, content: y, meta: {}}
        """
        raise NotImplementedError

    def get_revisions(self, uri):
        """
        Return list of tuples with uri and published state:
            [('i18n://sv-se@page/title.txt#1', False), ('i18n://sv-se@page/title.md#2', True)]
        """
        raise NotImplementedError


class DatabaseBackend(StorageBackend):

    def get(self, uri):
        node = self._get(uri)
        return self._serialize(uri, node)

    def get_many(self, uris):
        """
        Simple implementation,
        could be better implemented by backend not hitting db for every uri.
        """
        nodes = {}

        for uri in uris:
            try:
                node = self.get(uri)
            except NodeDoesNotExist:
                continue
            else:
                nodes[uri] = node

        return nodes

    def set(self, uri, content, **meta):
        """
        Dispatches private update/create handlers
        """
        try:
            node = self._update(uri, content, **meta)
            created = False
        except NodeDoesNotExist:
            node = self._create(uri, content, **meta)
            created = True
        return self._serialize(uri, node), created

    def delete(self, uri):
        node = None
        try:
            _node = self._get(uri)
        except NodeDoesNotExist:
            logger.warn('Tried to delete non existing node from storage: "%s"', uri)
        else:
            node = self._serialize(uri, _node)
            self._delete(_node)
        return node

    def delete_many(self, uris):
        """
        Simple implementation,
        could be better implemented by backend not hitting db for every uri.
        """
        deleted_nodes = {}

        for uri in uris:
            node = self.delete(uri)
            if node:
                deleted_nodes[uri] = node

        return deleted_nodes

    def _get(self, uri):
        raise NotImplementedError

    def _create(self, uri, content, **meta):
        raise NotImplementedError

    def _update(self, uri, content, **meta):
        raise NotImplementedError

    def _delete(self, node):
        raise NotImplementedError

    def _build_key(self, uri):
        """
        Build node identifying key for base uri.
        """
        return uri.clone(ext=None, version=None)

    def _serialize(self, uri, node):
        """
        Serialize node result as dict
        """
        meta = self._decode_meta(node['meta'], is_published=bool(node['is_published']))
        return {
            'uri': uri.clone(ext=node['plugin'], version=node['version']),
            'content': node['content'],
            'meta': meta
        }

    def _decode_meta(self, meta, **extra):
        """
        Decode and load underlying meta structure to dict and apply optional extra values.
        """
        _meta = json.loads(meta) if meta else {}
        _meta.update(extra)
        return _meta

    def _encode_meta(self, meta):
        """
        Encode meta dict for underlying persistence.
        """
        return json.dumps(meta) if meta else None

    def _merge_meta(self, encoded_meta, meta):
        """
        Merge new meta dict into encoded meta. Returns new encoded meta.
        """
        new_meta = None

        if meta:
            _meta = self._decode_meta(encoded_meta)
            for key, value in meta.iteritems():
                if value is None:
                    _meta.pop(key, None)
                else:
                    _meta[key] = value
            new_meta = self._encode_meta(_meta)

        return new_meta

    def _get_next_version(self, revisions):
        """
        Calculates new version number based on existing numeric ones.
        """
        versions = [0]
        for v in revisions:
            if v.isdigit():
                versions.append(int(v))
        return str(sorted(versions)[-1] + 1)



from threading import local


class ThreadLocalObject(local):
    """
    Base class to inherit from when thread local instances are needed.
    """
    initialized = False

    def __init__(self):
        if self.initialized:
            raise SystemError('%s initialized too many times' % self.__class__.__name__)
        self.initialized = True



from .base import BasePipe
from ...backends import cache


class CachePipe(BasePipe):

    def get_request(self, request):
        response = {}

        # Only get nodes from cache without specified version
        uris = tuple(uri for uri, node in request.iteritems() if not node.uri.version)

        if uris:
            cached_nodes = cache.get_many(uris)

            for uri, cached_node in cached_nodes.iteritems():
                node = response[node.uri] = request.pop(uri)
                self.materialize_node(node, **cached_node)

        return response

    def get_response(self, response):
        nodes = {}

        # Cache nodes without specified version (i.e. default or published)
        for uri, node in response.iteritems():
            if not uri.version:
                origin_uri = node.uri.clone(namespace=uri.namespace)
                nodes[origin_uri] = node.content
                # Empty node meta to be coherent with cached nodes
                node.meta.clear()

        if nodes:
            cache.set_many(nodes)

        return response

    def publish_response(self, response):
        nodes = dict((node.uri, node.content) for uri, node in response.iteritems())
        cache.set_many(nodes)
        return response

    def delete_response(self, response):
        cache.delete_many(response.keys())
        return response



from cio.conf import settings


class BasePlugin(object):

    ext = None

    @property
    def settings(self):
        return settings.get(self.ext.upper(), {})

    def load(self, content):
        """
        Return plugin data for content string
        """
        return content

    def save(self, data):
        """
        Persist external plugin resources and return content string for plugin data
        """
        return data

    def delete(self, data):
        """
        Delete external plugin resources
        """
        pass

    def render(self, data):
        """
        Render plugin
        """
        return data
